/*
 *  Copyright Droids Corporation (2011)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Revision : $Id: commands.c,v 1.9 2009-11-08 17:24:33 zer0 Exp $
 *
 *  Olivier MATZ <zer0@droids-corp.org>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <aversive.h>
#include <aversive/pgmspace.h>
#include <aversive/queue.h>
#include <aversive/endian.h>
#include <aversive/error.h>
#include <aversive/wait.h>
#include <parse.h>
#include <rdline.h>
#include <parse_string.h>
#include <parse_num.h>
#include <uart.h>
#include <xbee.h>

#include "callout.h"
#include "parse_atcmd.h"
#include "parse_neighbor.h"
#include "parse_monitor.h"

#include "spi_servo.h"
#include "rc_proto.h"
#include "main.h"
#include "cmdline.h"
#include "beep.h"
#include "eeprom_config.h"

/* commands_gen.c */
extern const parse_inst_t PROGMEM cmd_reset;
extern const parse_inst_t PROGMEM cmd_bootloader;
extern const parse_inst_t PROGMEM cmd_log;
extern const parse_inst_t PROGMEM cmd_log_show;
extern const parse_inst_t PROGMEM cmd_log_type;
extern const parse_inst_t PROGMEM cmd_stack_space;
extern const parse_inst_t PROGMEM cmd_scheduler;

static int monitor_period_ms = 1000;
static int monitor_running = 0;
static int monitor_count = 0;
static struct callout monitor_event;
struct monitor_reg *monitor_current;

static int range_period_ms = 1000;
static int range_powermask = 0x1F;
//static uint8_t range_power = 0;
static int range_running = 0;
static uint64_t range_dstaddr = 0xFFFF; /* broadcast by default */
static struct callout range_event;
static int range_count = 100;
static int range_cur_count = 0;

static void monitor_cb(struct callout_manager *cm,
		       struct callout *clt, void *dummy)
{
	(void)clt;
	(void)dummy;

	if (monitor_current == NULL)
		monitor_current = LIST_FIRST(&xbee_monitor_list);

	xbeeapp_send_atcmd(monitor_current->atcmd, NULL, 0, 0, NULL, NULL);
	monitor_current = LIST_NEXT(monitor_current, next);
	callout_reset(cm, &monitor_event,
		      monitor_period_ms / monitor_count,
		      SINGLE, monitor_cb, NULL);
}

static void range_cb(struct callout_manager *cm,
		     struct callout *clt, void *dummy)
{
	(void)cm;
	(void)clt;
	(void)dummy;
#if 0
	uint8_t i, mask;
	struct rc_proto_range rangepkt;

	(void)clt;
	(void)dummy;

	range_cur_count--;

	/* get new xmit power */
	for (i = 1; i <= 8; i++) {
		mask = 1 << ((range_power + i) & 0x7);
		if (mask & range_powermask)
			break;
	}
	range_power = ((range_power + i) & 0x7);

	xbeeapp_send_atcmd("PL", &range_power, sizeof(range_power), 0, NULL, NULL);

	rangepkt.type = RC_PROTO_TYPE_RANGE;
	rangepkt.power_level = range_power;

	xbeeapp_send_msg(range_dstaddr, &rangepkt, sizeof(rangepkt), 0);

	if (range_cur_count == 0) {
		range_running = 0;
		return;
	}

	callout_reset(cm, &range_event,
		      range_period_ms,
		      SINGLE, range_cb, NULL);
#endif
}

/* this structure is filled when cmd_help is parsed successfully */
struct cmd_help_result {
	fixed_string_t help;
	struct xbee_atcmd *cmd;
};

/* function called when cmd_help is parsed successfully */
static void cmd_help_parsed(void *parsed_result, void *data)
{
	struct cmd_help_result *res = parsed_result;
	struct xbee_atcmd cmdcopy;
	int type;

	(void)data;

	memcpy_P(&cmdcopy, res->cmd, sizeof(cmdcopy));
	type = (cmdcopy.flags & (XBEE_ATCMD_F_READ | XBEE_ATCMD_F_WRITE));
	switch (type) {
		case XBEE_ATCMD_F_READ:
			printf_P(PSTR("Read-only\r\n"));
			break;
		case XBEE_ATCMD_F_WRITE:
			printf_P(PSTR("Write-only\r\n"));
			break;
		default:
			printf_P(PSTR("Read-write\r\n"));
			break;
	}
	if (cmdcopy.flags & XBEE_ATCMD_F_PARAM_NONE)
		printf_P(PSTR("No argument\r\n"));
	else if (cmdcopy.flags & XBEE_ATCMD_F_PARAM_U8)
		printf_P(PSTR("Register is unsigned 8 bits\r\n"));
	else if (cmdcopy.flags & XBEE_ATCMD_F_PARAM_U16)
		printf_P(PSTR("Register is unsigned 16 bits\r\n"));
	else if (cmdcopy.flags & XBEE_ATCMD_F_PARAM_U32)
		printf_P(PSTR("Register is unsigned 32 bits\r\n"));
	else if (cmdcopy.flags & XBEE_ATCMD_F_PARAM_S16)
		printf_P(PSTR("Register is signed 16 bits\r\n"));
	else if (cmdcopy.flags & XBEE_ATCMD_F_PARAM_STRING_20B)
		printf_P(PSTR("Register is a 20 bytes string\r\n"));
	else
		printf_P(PSTR("Unknown argument\r\n"));

	printf_P(PSTR("%S\r\n"), cmdcopy.help);
}
const char PROGMEM str_help_help[] = "help";

const parse_token_string_t PROGMEM cmd_help_help =
	TOKEN_STRING_INITIALIZER(struct cmd_help_result, help, str_help_help);

const parse_token_atcmd_t PROGMEM cmd_help_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_help_result, cmd, &xbee_dev,
				0, 0);

const char PROGMEM help_help[] = "Help a register using an AT command";
const parse_inst_t PROGMEM cmd_help = {
	.f = cmd_help_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_help,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_help_help,
		(PGM_P)&cmd_help_atcmd,
		NULL,
	},
};

/* ************* */

struct cmd_neigh_del_result {
	fixed_string_t cmd;
	fixed_string_t action;
	struct xbee_neigh *neigh;
};

static void cmd_neigh_del_parsed(void *parsed_result,
				void *data)
{
	struct cmd_neigh_del_result *res = parsed_result;

	(void)data;
	xbee_neigh_del(xbee_dev, res->neigh);
}

const char PROGMEM str_neigh_del_neigh[] = "neigh";
const parse_token_string_t PROGMEM cmd_neigh_del_cmd =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_del_result, cmd,
				 str_neigh_del_neigh);
const char PROGMEM str_neigh_del_del[] = "del";
const parse_token_string_t PROGMEM cmd_neigh_del_action =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_del_result, action,
				 str_neigh_del_del);
const parse_token_neighbor_t PROGMEM cmd_neigh_del_neigh =
	TOKEN_NEIGHBOR_INITIALIZER(struct cmd_neigh_del_result, neigh,
				   &xbee_dev);

const char PROGMEM help_neigh_del[] = "delete a neighbor";
const parse_inst_t PROGMEM cmd_neigh_del = {
	.f = cmd_neigh_del_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_neigh_del,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_neigh_del_cmd,
		(PGM_P)&cmd_neigh_del_action,
		(PGM_P)&cmd_neigh_del_neigh,
		NULL,
	},
};

/* ************* */

struct cmd_neigh_add_result {
	fixed_string_t cmd;
	fixed_string_t action;
	fixed_string_t name;
	uint64_t addr;
};

static void cmd_neigh_add_parsed(void *parsed_result,
				 void *data)
{
	struct cmd_neigh_add_result *res = parsed_result;

	(void)data;
	if (xbee_neigh_add(xbee_dev, res->name, res->addr) == NULL)
		printf_P(PSTR("name or addr already exist\r\n"));
}

const char PROGMEM str_neigh_add_neigh[] = "neigh";
const parse_token_string_t PROGMEM cmd_neigh_add_cmd =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_add_result, cmd,
				 str_neigh_add_neigh);
const char PROGMEM str_neigh_add_add[] = "add";
const parse_token_string_t PROGMEM cmd_neigh_add_action =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_add_result, action,
				 str_neigh_add_add);
const parse_token_string_t PROGMEM cmd_neigh_add_name =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_add_result, name, NULL);
const parse_token_num_t PROGMEM cmd_neigh_add_addr =
	TOKEN_NUM_INITIALIZER(struct cmd_neigh_add_result, addr, UINT64);

const char PROGMEM help_neigh_add[] = "add a neighbor";
const parse_inst_t PROGMEM cmd_neigh_add = {
	.f = cmd_neigh_add_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_neigh_add,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_neigh_add_cmd,
		(PGM_P)&cmd_neigh_add_action,
		(PGM_P)&cmd_neigh_add_name,
		(PGM_P)&cmd_neigh_add_addr,
		NULL,
	},
};

/* ************* */

struct cmd_neigh_list_result {
	fixed_string_t cmd;
	fixed_string_t action;
};

static void cmd_neigh_list_parsed(void *parsed_result,
				void *data)
{
	struct xbee_neigh *neigh;

	(void)parsed_result;
	(void)data;
	LIST_FOREACH(neigh, &xbee_dev->neigh_list, next) {
		printf_P(PSTR(" %s: 0x%.8"PRIx32"%.8"PRIx32"\r\n"),
			 neigh->name,
			 (uint32_t)(neigh->addr >> 32ULL),
			 (uint32_t)(neigh->addr & 0xFFFFFFFF));
	}
}

const char PROGMEM str_neigh_list_neigh[] = "neigh";
const parse_token_string_t PROGMEM cmd_neigh_list_cmd =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_list_result, cmd,
				 str_neigh_list_neigh);
const char PROGMEM str_neigh_list_list[] = "list";
const parse_token_string_t PROGMEM cmd_neigh_list_action =
	TOKEN_STRING_INITIALIZER(struct cmd_neigh_list_result, action,
				 str_neigh_list_list);

const char PROGMEM help_neigh_list[] = "list all knwon neighbors";
const parse_inst_t PROGMEM cmd_neigh_list = {
	.f = cmd_neigh_list_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_neigh_list,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_neigh_list_cmd,
		(PGM_P)&cmd_neigh_list_action,
		NULL,
	},
};




/* ************* */

/* this structure is filled when cmd_read is parsed successfully */
struct cmd_read_result {
	fixed_string_t read;
	struct xbee_atcmd *cmd;
};

/* function called when cmd_read is parsed successfully */
static void cmd_read_parsed(void *parsed_result,
			    void *data)
{
	struct cmd_read_result *res = parsed_result;
	struct xbee_atcmd copy;
	char cmd[3];

	(void)data;
	memcpy_P(&copy, res->cmd, sizeof(copy));
	memcpy_P(&cmd, copy.name, 2);
	cmd[2] = '\0';
	xbeeapp_send_atcmd(cmd, NULL, 0, 1, NULL, NULL);
}

const char PROGMEM str_read_read[] = "read";

const parse_token_string_t PROGMEM cmd_read_read =
	TOKEN_STRING_INITIALIZER(struct cmd_read_result, read,
				 str_read_read);

const parse_token_atcmd_t PROGMEM cmd_read_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_read_result, cmd, &xbee_dev,
				XBEE_ATCMD_F_READ, XBEE_ATCMD_F_READ);

const char PROGMEM help_read[] = "Read a register using an AT command";
const parse_inst_t PROGMEM cmd_read = {
	.f = cmd_read_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_read,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_read_read,
		(PGM_P)&cmd_read_atcmd,
		NULL,
	},
};


/* ************* */

/* this structure is filled when cmd_write is parsed successfully */
struct cmd_write_result {
	fixed_string_t write;
	struct xbee_atcmd *cmd;
	union {
		uint8_t u8;
		uint16_t u16;
		uint32_t u32;
	};
};

/* function called when cmd_write is parsed successfully */
static void cmd_write_parsed(void *parsed_result, void *data)
{
	struct cmd_write_result *res = parsed_result;
	struct xbee_atcmd copy;
	char cmd[3];
	int len;
	void *param;

	(void)data;
	memcpy_P(&copy, res->cmd, sizeof(copy));

	if (copy.flags & XBEE_ATCMD_F_PARAM_NONE) {
		len = 0;
		param = NULL;
	}
	else if (copy.flags & XBEE_ATCMD_F_PARAM_U8) {
		len = sizeof(res->u8);
		param = &res->u8;
	}
	else if (copy.flags & XBEE_ATCMD_F_PARAM_U16) {
		len = sizeof(res->u16);
		res->u16 = htons(res->u16);
		param = &res->u16;
	}
	else if (copy.flags & XBEE_ATCMD_F_PARAM_U32) {
		len = sizeof(res->u32);
		res->u32 = htonl(res->u32);
		param = &res->u32;
	}
	else {
		printf_P(PSTR("Unknown argument type\r\n"));
		return;
	}
	memcpy_P(&cmd, copy.name, 2);
	cmd[2] = '\0';
	xbeeapp_send_atcmd(cmd, param, len, 1, NULL, NULL);
}

const char PROGMEM str_write_none[] = "write";

const parse_token_string_t PROGMEM cmd_write_write =
	TOKEN_STRING_INITIALIZER(struct cmd_write_result, write,
				 str_write_none);

const parse_token_atcmd_t PROGMEM cmd_write_none_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_write_result, cmd,
				&xbee_dev,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_NONE,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_NONE);

const char PROGMEM help_write_none[] = "Send an AT command (no argument)";

const parse_inst_t PROGMEM cmd_write_none = {
	.f = cmd_write_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_write_none,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_write_write,
		(PGM_P)&cmd_write_none_atcmd,
		NULL,
	},
};

const parse_token_atcmd_t PROGMEM cmd_write_u8_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_write_result, cmd,
				&xbee_dev,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_U8,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_U8);

const parse_token_num_t PROGMEM cmd_write_u8_u8 =
	TOKEN_NUM_INITIALIZER(struct cmd_write_result, u8, UINT8);

const char PROGMEM help_write_u8[] = "Write a 8 bits register using an AT command";

const parse_inst_t PROGMEM cmd_write_u8 = {
	.f = cmd_write_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_write_u8,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_write_write,
		(PGM_P)&cmd_write_u8_atcmd,
		(PGM_P)&cmd_write_u8_u8,
		NULL,
	},
};

const parse_token_atcmd_t PROGMEM cmd_write_u16_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_write_result, cmd,
				&xbee_dev,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_U16,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_U16);

const parse_token_num_t PROGMEM cmd_write_u16_u16 =
	TOKEN_NUM_INITIALIZER(struct cmd_write_result, u16, UINT16);

const char PROGMEM help_write_u16[] = "Write a 16 bits register using an AT command";

const parse_inst_t PROGMEM cmd_write_u16 = {
	.f = cmd_write_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_write_u16,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_write_write,
		(PGM_P)&cmd_write_u16_atcmd,
		(PGM_P)&cmd_write_u16_u16,
		NULL,
	},
};

const parse_token_atcmd_t PROGMEM cmd_write_u32_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_write_result, cmd,
				&xbee_dev,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_U32,
				XBEE_ATCMD_F_WRITE | XBEE_ATCMD_F_PARAM_U32);

const parse_token_num_t PROGMEM cmd_write_u32_u32 =
	TOKEN_NUM_INITIALIZER(struct cmd_write_result, u32, UINT32);

const char PROGMEM help_write_u32[] = "Write a 32 bits register using an AT command";

const parse_inst_t PROGMEM cmd_write_u32 = {
	.f = cmd_write_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_write_u32,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_write_write,
		(PGM_P)&cmd_write_u32_atcmd,
		(PGM_P)&cmd_write_u32_u32,
		NULL,
	},
};


/* ************* */

/* this structure is filled when cmd_sendmsg is parsed successfully */
struct cmd_sendmsg_result {
	fixed_string_t sendmsg;
	uint64_t addr;
	fixed_string_t data;
};

/* function called when cmd_sendmsg is parsed successfully */
static void cmd_sendmsg_parsed(void *parsed_result, void *data)
{
	struct cmd_sendmsg_result *res = parsed_result;

	(void)data;
	xbeeapp_send_msg(res->addr, res->data, strlen(res->data), 1);
}

const char PROGMEM str_sendmsg[] = "sendmsg";

const parse_token_string_t PROGMEM cmd_sendmsg_sendmsg =
	TOKEN_STRING_INITIALIZER(struct cmd_sendmsg_result, sendmsg,
				 str_sendmsg);

const parse_token_num_t PROGMEM cmd_sendmsg_addr =
	TOKEN_NUM_INITIALIZER(struct cmd_sendmsg_result, addr, UINT64);

const parse_token_string_t PROGMEM cmd_sendmsg_data =
	TOKEN_STRING_INITIALIZER(struct cmd_sendmsg_result, data, NULL);

const char PROGMEM help_sendmsg[] = "Send data to a node using its address";

const parse_inst_t PROGMEM cmd_sendmsg = {
	.f = cmd_sendmsg_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_sendmsg,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_sendmsg_sendmsg,
		(PGM_P)&cmd_sendmsg_addr,
		(PGM_P)&cmd_sendmsg_data,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_send_hello is parsed successfully */
struct cmd_send_hello_result {
	fixed_string_t send_hello;
	uint64_t addr;
	struct xbee_neigh *neigh;
	uint16_t period;
	uint16_t count;
	fixed_string_t data;
};

/* function called when cmd_send_hello is parsed successfully */
static void cmd_send_hello_parsed(void *parsed_result, void *use_neigh)
{
	struct cmd_send_hello_result *res = parsed_result;
	uint16_t now, next, diff;
	uint8_t flags;
	uint64_t addr;

	if (use_neigh)
		addr = res->neigh->addr;
	else
		addr = res->addr;

	IRQ_LOCK(flags);
	now = global_ms;
	IRQ_UNLOCK(flags);

	next = now;

	while (!cmdline_keypressed() && res->count != 0) {
		IRQ_LOCK(flags);
		now = global_ms;
		IRQ_UNLOCK(flags);

		diff = now - next;
		if (diff < res->period)
			continue;

		rc_proto_send_hello(addr, res->data, strlen(res->data));
		next += res->period;
		res->count--;
	}
}

const char PROGMEM str_send_hello[] = "send_hello";

const parse_token_string_t PROGMEM cmd_send_hello_send_hello =
	TOKEN_STRING_INITIALIZER(struct cmd_send_hello_result, send_hello,
				 str_send_hello);

const parse_token_num_t PROGMEM cmd_send_hello_addr =
	TOKEN_NUM_INITIALIZER(struct cmd_send_hello_result, addr, UINT64);

const parse_token_num_t PROGMEM cmd_send_hello_period =
	TOKEN_NUM_INITIALIZER(struct cmd_send_hello_result, period, UINT16);

const parse_token_num_t PROGMEM cmd_send_hello_count =
	TOKEN_NUM_INITIALIZER(struct cmd_send_hello_result, count, UINT16);

const parse_token_string_t PROGMEM cmd_send_hello_data =
	TOKEN_STRING_INITIALIZER(struct cmd_send_hello_result, data, NULL);

const char PROGMEM help_send_hello[] =
	"Send hello msg to a node: addr, period_ms, count, str";

const parse_inst_t PROGMEM cmd_send_hello = {
	.f = cmd_send_hello_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_send_hello,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_send_hello_send_hello,
		(PGM_P)&cmd_send_hello_addr,
		(PGM_P)&cmd_send_hello_period,
		(PGM_P)&cmd_send_hello_count,
		(PGM_P)&cmd_send_hello_data,
		NULL,
	},
};

const parse_token_neighbor_t PROGMEM cmd_send_hello_neigh =
	TOKEN_NEIGHBOR_INITIALIZER(struct cmd_send_hello_result, neigh,
				   &xbee_dev);

const parse_inst_t PROGMEM cmd_send_hello_name = {
	.f = cmd_send_hello_parsed,  /* function to call */
	.data = (void *)1,      /* 2nd arg of func */
	.help_str = help_send_hello,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_send_hello_send_hello,
		(PGM_P)&cmd_send_hello_neigh,
		(PGM_P)&cmd_send_hello_period,
		(PGM_P)&cmd_send_hello_count,
		(PGM_P)&cmd_send_hello_data,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_sendmsg_name is parsed successfully */
struct cmd_sendmsg_name_result {
	fixed_string_t sendmsg_name;
	struct xbee_neigh *neigh;
	fixed_string_t data;
};

/* function called when cmd_sendmsg_name is parsed successfully */
static void cmd_sendmsg_name_parsed(void *parsed_result, void *data)
{
	struct cmd_sendmsg_name_result *res = parsed_result;

	(void)data;
	xbeeapp_send_msg(res->neigh->addr, res->data, strlen(res->data), 1);
}

const parse_token_string_t PROGMEM cmd_sendmsg_name_sendmsg_name =
	TOKEN_STRING_INITIALIZER(struct cmd_sendmsg_name_result, sendmsg_name,
				 str_sendmsg);

const parse_token_neighbor_t PROGMEM cmd_sendmsg_name_neigh =
	TOKEN_NEIGHBOR_INITIALIZER(struct cmd_sendmsg_name_result, neigh,
				   &xbee_dev);

const parse_token_string_t PROGMEM cmd_sendmsg_name_data =
	TOKEN_STRING_INITIALIZER(struct cmd_sendmsg_name_result, data, NULL);

const char PROGMEM help_sendmsg_name[] = "Send data to a node using its name";

const parse_inst_t PROGMEM cmd_sendmsg_name = {
	.f = cmd_sendmsg_name_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_sendmsg_name,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_sendmsg_name_sendmsg_name,
		(PGM_P)&cmd_sendmsg_name_neigh,
		(PGM_P)&cmd_sendmsg_name_data,
		NULL,
	},
};


/* ************* */

/* this structure is filled when cmd_range is parsed successfully */
struct cmd_range_result {
	fixed_string_t range;
	fixed_string_t action;
};

/* function called when cmd_range is parsed successfully */
static void cmd_range_parsed(void *parsed_result, void *data)
{
	struct cmd_range_result *res = parsed_result;

	(void)data;
	if (!strcmp_P(res->action, PSTR("show"))) {
		printf_P(PSTR("range infos:\r\n"));
		printf_P(PSTR("  range period %d\r\n"), range_period_ms);
		printf_P(PSTR("  range count %d\r\n"), range_count);
		printf_P(PSTR("  range powermask 0x%x\r\n"), range_powermask);
		printf_P(PSTR("  range dstaddr 0x%.8"PRIx32"%.8"PRIx32"\r\n"),
			 (uint32_t)(range_dstaddr >> 32ULL),
			 (uint32_t)(range_dstaddr & 0xFFFFFFFF));

		if (range_running)
			printf_P(PSTR("  range test is running\r\n"));
		else
			printf_P(PSTR("  range test is not running\r\n"));
	}
	else if (!strcmp(res->action, "start")) {
		if (range_running) {
			printf_P(PSTR("already running\r\n"));
			return;
		}
		range_cur_count = range_count;
		callout_init(&range_event);
		callout_reset(&cm, &range_event, 0,
			      SINGLE, range_cb, NULL);
		range_running = 1;
	}
	else if (!strcmp(res->action, "end")) {
		if (range_running == 0) {
			printf_P(PSTR("not running\r\n"));
			return;
		}
		range_running = 0;
		callout_stop(&cm, &range_event);
	}
}

const char PROGMEM str_range[] = "range";
const char PROGMEM str_range_tokens[] = "show#start#end";

const parse_token_string_t PROGMEM cmd_range_range =
	TOKEN_STRING_INITIALIZER(struct cmd_range_result, range,
				 str_range);
const parse_token_string_t PROGMEM cmd_range_action =
	TOKEN_STRING_INITIALIZER(struct cmd_range_result, action,
				 str_range_tokens);

const char PROGMEM help_range[] = "start/stop/show current rangeing";

const parse_inst_t PROGMEM cmd_range = {
	.f = cmd_range_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_range,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_range_range,
		(PGM_P)&cmd_range_action,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_range_period is parsed successfully */
struct cmd_range_period_result {
	fixed_string_t range;
	fixed_string_t action;
	uint32_t period;
};

/* function called when cmd_range_period is parsed successfully */
static void cmd_range_period_parsed(void *parsed_result, void *data)
{
	struct cmd_range_period_result *res = parsed_result;

	(void)data;
	if (res->period < 10) {
		printf_P(PSTR("error, minimum period is 10 ms\r\n"));
		return;
	}

	range_period_ms = res->period;
}

const char PROGMEM str_period[] = "period";

const parse_token_string_t PROGMEM cmd_range_period_range_period =
	TOKEN_STRING_INITIALIZER(struct cmd_range_period_result, range,
				 str_range);
const parse_token_string_t PROGMEM cmd_range_period_action =
	TOKEN_STRING_INITIALIZER(struct cmd_range_period_result, action,
				 str_period);
const parse_token_num_t PROGMEM cmd_range_period_period =
	TOKEN_NUM_INITIALIZER(struct cmd_range_period_result, period, UINT32);

const char PROGMEM help_range_period[] = "set range test period";

const parse_inst_t PROGMEM cmd_range_period = {
	.f = cmd_range_period_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_range_period,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_range_period_range_period,
		(PGM_P)&cmd_range_period_action,
		(PGM_P)&cmd_range_period_period,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_range_count is parsed successfully */
struct cmd_range_count_result {
	fixed_string_t range;
	fixed_string_t action;
	uint32_t count;
};

/* function called when cmd_range_count is parsed successfully */
static void cmd_range_count_parsed(void *parsed_result, void *data)
{
	struct cmd_range_count_result *res = parsed_result;

	(void)data;
	range_count = res->count;
}

const char PROGMEM str_count[] = "count";

const parse_token_string_t PROGMEM cmd_range_count_range_count =
	TOKEN_STRING_INITIALIZER(struct cmd_range_count_result, range,
				 str_range);
const parse_token_string_t PROGMEM cmd_range_count_action =
	TOKEN_STRING_INITIALIZER(struct cmd_range_count_result, action,
				 str_count);
const parse_token_num_t PROGMEM cmd_range_count_count =
	TOKEN_NUM_INITIALIZER(struct cmd_range_count_result, count, UINT32);


const char PROGMEM help_range_count[] = "set range test count";

const parse_inst_t PROGMEM cmd_range_count = {
	.f = cmd_range_count_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_range_count,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_range_count_range_count,
		(PGM_P)&cmd_range_count_action,
		(PGM_P)&cmd_range_count_count,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_range_powermask is parsed successfully */
struct cmd_range_powermask_result {
	fixed_string_t range;
	fixed_string_t action;
	uint8_t powermask;
};

/* function called when cmd_range_powermask is parsed successfully */
static void cmd_range_powermask_parsed(void *parsed_result, void *data)
{
	struct cmd_range_powermask_result *res = parsed_result;

	(void)data;
	range_powermask = res->powermask;
}

const char PROGMEM str_powermask[] = "powermask";

const parse_token_string_t PROGMEM cmd_range_powermask_range_powermask =
	TOKEN_STRING_INITIALIZER(struct cmd_range_powermask_result, range,
				 str_range);
const parse_token_string_t PROGMEM cmd_range_powermask_action =
	TOKEN_STRING_INITIALIZER(struct cmd_range_powermask_result, action,
				 str_powermask);
const parse_token_num_t PROGMEM cmd_range_powermask_powermask =
	TOKEN_NUM_INITIALIZER(struct cmd_range_powermask_result, powermask,
			      UINT8);


const char PROGMEM help_range_powermask[] = "set range test powermask";

const parse_inst_t PROGMEM cmd_range_powermask = {
	.f = cmd_range_powermask_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_range_powermask,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_range_powermask_range_powermask,
		(PGM_P)&cmd_range_powermask_action,
		(PGM_P)&cmd_range_powermask_powermask,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_range_dstaddr is parsed successfully */
struct cmd_range_dstaddr_result {
	fixed_string_t range;
	fixed_string_t action;
	uint64_t dstaddr;
};

/* function called when cmd_range_dstaddr is parsed successfully */
static void cmd_range_dstaddr_parsed(void *parsed_result, void *data)
{
	struct cmd_range_dstaddr_result *res = parsed_result;

	(void)data;
	range_dstaddr = res->dstaddr;
}

const char PROGMEM str_dstaddr[] = "dstaddr";

const parse_token_string_t PROGMEM cmd_range_dstaddr_range_dstaddr =
	TOKEN_STRING_INITIALIZER(struct cmd_range_dstaddr_result, range,
				 str_range);
const parse_token_string_t PROGMEM cmd_range_dstaddr_action =
	TOKEN_STRING_INITIALIZER(struct cmd_range_dstaddr_result, action,
				 str_dstaddr);
const parse_token_num_t PROGMEM cmd_range_dstaddr_dstaddr =
	TOKEN_NUM_INITIALIZER(struct cmd_range_dstaddr_result, dstaddr, UINT64);


const char PROGMEM help_range_dstaddr[] = "set register rangeing dstaddr";

const parse_inst_t PROGMEM cmd_range_dstaddr = {
	.f = cmd_range_dstaddr_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_range_dstaddr,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_range_dstaddr_range_dstaddr,
		(PGM_P)&cmd_range_dstaddr_action,
		(PGM_P)&cmd_range_dstaddr_dstaddr,
		NULL,
	},
};


/* ************* */

/* this structure is filled when cmd_monitor is parsed successfully */
struct cmd_monitor_result {
	fixed_string_t monitor;
	fixed_string_t action;
};

/* function called when cmd_monitor is parsed successfully */
static void cmd_monitor_parsed(void *parsed_result, void *data)
{
	struct cmd_monitor_result *res = parsed_result;
	struct monitor_reg *m;

	(void)data;
	if (!strcmp_P(res->action, PSTR("show"))) {
		printf_P(PSTR("monitor period is %d ms, %d regs in list\r\n"),
		       monitor_period_ms, monitor_count);
		LIST_FOREACH(m, &xbee_monitor_list, next)
			printf_P(PSTR(" %S\r\n"), m->desc);
	}
	else if (!strcmp_P(res->action, PSTR("start"))) {
		if (monitor_running) {
			printf_P(PSTR("already running\r\n"));
			return;
		}
		if (monitor_count == 0) {
			printf_P(PSTR("no regs to be monitored\r\n"));
			return;
		}
		callout_init(&monitor_event);
		callout_reset(&cm, &monitor_event, 0, SINGLE, monitor_cb, NULL);
		monitor_running = 1;
		monitor_current = LIST_FIRST(&xbee_monitor_list);
		printf_P(PSTR("monitor cb: %S %s\r\n"),
			 monitor_current->desc,
			 monitor_current->atcmd);

	}
	else if (!strcmp_P(res->action, PSTR("end"))) {
		if (monitor_running == 0) {
			printf_P(PSTR("not running\r\n"));
			return;
		}
		monitor_running = 0;
		callout_stop(&cm, &monitor_event);
	}
}

const char PROGMEM str_monitor[] = "monitor";
const char PROGMEM str_monitor_tokens[] = "show#start#end";

const parse_token_string_t PROGMEM cmd_monitor_monitor =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_result, monitor,
				 str_monitor);
const parse_token_string_t PROGMEM cmd_monitor_action =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_result, action,
				 str_monitor_tokens);

const char PROGMEM help_monitor[] = "start/stop/show current monitoring";

const parse_inst_t PROGMEM cmd_monitor = {
	.f = cmd_monitor_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_monitor,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_monitor_monitor,
		(PGM_P)&cmd_monitor_action,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_monitor_add is parsed successfully */
struct cmd_monitor_add_result {
	fixed_string_t monitor;
	fixed_string_t action;
	struct xbee_atcmd *cmd;
};

/* function called when cmd_monitor_add is parsed successfully */
static void cmd_monitor_add_parsed(void *parsed_result, void *data)
{
	struct cmd_monitor_add_result *res = parsed_result;
	struct monitor_reg *m;
	struct xbee_atcmd copy;

	(void)data;
	memcpy_P(&copy, res->cmd, sizeof(copy));
	LIST_FOREACH(m, &xbee_monitor_list, next) {
		if (!strcmp_P(m->atcmd, copy.name))
			break;
	}

	if (m != NULL) {
		printf_P(PSTR("already exist\r\n"));
		return;
	}

	m = malloc(sizeof(*m));
	if (m == NULL) {
		printf_P(PSTR("no mem\r\n"));
		return;
	}
	m->desc = copy.desc;
	strcpy_P(m->atcmd, copy.name);
	LIST_INSERT_HEAD(&xbee_monitor_list, m, next);
	monitor_count ++;
}

const char PROGMEM str_monitor_add[] = "add";

const parse_token_string_t PROGMEM cmd_monitor_add_monitor_add =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_add_result, monitor,
				 str_monitor);
const parse_token_string_t PROGMEM cmd_monitor_add_action =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_add_result, action,
				 str_monitor_add);
const parse_token_atcmd_t PROGMEM cmd_monitor_add_atcmd =
	TOKEN_ATCMD_INITIALIZER(struct cmd_monitor_add_result, cmd, &xbee_dev,
				XBEE_ATCMD_F_READ, XBEE_ATCMD_F_READ);


const char PROGMEM help_monitor_add[] = "add a register in monitor list";

const parse_inst_t PROGMEM cmd_monitor_add = {
	.f = cmd_monitor_add_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_monitor_add,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_monitor_add_monitor_add,
		(PGM_P)&cmd_monitor_add_action,
		(PGM_P)&cmd_monitor_add_atcmd,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_monitor_period is parsed successfully */
struct cmd_monitor_period_result {
	fixed_string_t monitor;
	fixed_string_t action;
	uint32_t period;
};

/* function called when cmd_monitor_period is parsed successfully */
static void cmd_monitor_period_parsed(void *parsed_result, void *data)
{
	struct cmd_monitor_period_result *res = parsed_result;

	(void)data;
	if (res->period < 100) {
		printf_P(PSTR("error, minimum period is 100 ms\r\n"));
		return;
	}

	monitor_period_ms = res->period;
}

const char PROGMEM str_monitor_period[] = "period";

const parse_token_string_t PROGMEM cmd_monitor_period_monitor_period =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_period_result, monitor,
				 str_monitor);
const parse_token_string_t PROGMEM cmd_monitor_period_action =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_period_result, action,
				 str_monitor_period);
const parse_token_num_t PROGMEM cmd_monitor_period_period =
	TOKEN_NUM_INITIALIZER(struct cmd_monitor_period_result, period, UINT32);


const char PROGMEM help_monitor_period[] = "set register monitoring period";

const parse_inst_t PROGMEM cmd_monitor_period = {
	.f = cmd_monitor_period_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_monitor_period,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_monitor_period_monitor_period,
		(PGM_P)&cmd_monitor_period_action,
		(PGM_P)&cmd_monitor_period_period,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_monitor_del is parsed successfully */
struct cmd_monitor_del_result {
	fixed_string_t monitor;
	fixed_string_t action;
	struct monitor_reg *m;
};

/* function called when cmd_monitor_del is parsed successfully */
static void cmd_monitor_del_parsed(void *parsed_result, void *data)
{
	struct cmd_monitor_del_result *res = parsed_result;

	(void)data;
	monitor_current = LIST_NEXT(res->m, next);
	LIST_REMOVE(res->m, next);
	free(res->m);
	monitor_count --;
	if (monitor_count == 0) {
		printf_P(PSTR("Disable monitoring, no more event\r\n"));
		callout_stop(&cm, &monitor_event);
		monitor_running = 0;
		return;
	}
}

const char PROGMEM str_monitor_del[] = "del";

const parse_token_string_t PROGMEM cmd_monitor_del_monitor_del =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_del_result, monitor,
				 str_monitor);
const parse_token_string_t PROGMEM cmd_monitor_del_action =
	TOKEN_STRING_INITIALIZER(struct cmd_monitor_del_result, action,
				 str_monitor_del);
const parse_token_monitor_t PROGMEM cmd_monitor_del_atcmd =
	TOKEN_MONITOR_INITIALIZER(struct cmd_monitor_del_result, m);


const char PROGMEM help_monitor_del[] = "del a register in monitor list";

const parse_inst_t PROGMEM cmd_monitor_del = {
	.f = cmd_monitor_del_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_monitor_del,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_monitor_del_monitor_del,
		(PGM_P)&cmd_monitor_del_action,
		(PGM_P)&cmd_monitor_del_atcmd,
		NULL,
	},
};


/* ************* */

/* this structure is filled when cmd_ping is parsed successfully */
struct cmd_ping_result {
	fixed_string_t ping;
};

/* function called when cmd_ping is parsed successfully */
static void cmd_ping_parsed(void *parsed_result, void *data)
{
	(void)parsed_result;
	(void)data;
	xbeeapp_send_atcmd("VL", NULL, 0, 1, NULL, NULL);
}

const char PROGMEM str_ping[] = "ping";

const parse_token_string_t PROGMEM cmd_ping_ping =
	TOKEN_STRING_INITIALIZER(struct cmd_ping_result, ping,
				 str_ping);

const char PROGMEM help_ping[] = "Send a ping to the xbee device";

const parse_inst_t PROGMEM cmd_ping = {
	.f = cmd_ping_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_ping,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_ping_ping,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_raw is parsed successfully */
struct cmd_raw_result {
	fixed_string_t raw;
};

/* function called when cmd_raw is parsed successfully */
static void cmd_raw_parsed(void *parsed_result, void *data)
{
	(void)parsed_result;
	(void)data;
	printf_P(PSTR("switched to raw mode, CTRL-D to exit\r\n"));
	rdline_stop(&xbeeboard.rdl); /* don't display prompt when return */
	xbee_raw = 1;
}

const char PROGMEM str_raw[] = "raw";

const parse_token_string_t PROGMEM cmd_raw_raw =
	TOKEN_STRING_INITIALIZER(struct cmd_raw_result, raw,
				 str_raw);

const char PROGMEM help_raw[] = "Switch to raw mode";

const parse_inst_t PROGMEM cmd_raw = {
	.f = cmd_raw_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_raw,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_raw_raw,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_dump is parsed successfully */
struct cmd_dump_result {
	fixed_string_t dump;
	fixed_string_t onoff;
};

/* function called when cmd_dump is parsed successfully */
static void cmd_dump_parsed(void *parsed_result, void *data)
{
	struct cmd_dump_result *res = parsed_result;

	(void)data;
	if (!strcmp(res->onoff, "on"))
		xbee_hexdump = 1;
	else
		xbee_hexdump = 0;
}

const char PROGMEM str_dump[] = "dump";
const char PROGMEM str_dump_onoff[] = "on#off";

const parse_token_string_t PROGMEM cmd_dump_dump =
	TOKEN_STRING_INITIALIZER(struct cmd_dump_result, dump,
				 str_dump);

const parse_token_string_t PROGMEM cmd_dump_onoff =
	TOKEN_STRING_INITIALIZER(struct cmd_dump_result, onoff,
				 str_dump_onoff);

const char PROGMEM help_dump[] = "enable/disable hexdump of received packets";

const parse_inst_t PROGMEM cmd_dump = {
	.f = cmd_dump_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_dump,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_dump_dump,
		(PGM_P)&cmd_dump_onoff,
		NULL,
	},
};

/* ************* */

/* this structure is filled when cmd_debug is parsed successfully */
struct cmd_debug_result {
	fixed_string_t debug;
	fixed_string_t onoff;
};

/* function called when cmd_debug is parsed successfully */
static void cmd_debug_parsed(void *parsed_result, void *data)
{
	struct cmd_debug_result *res = parsed_result;

	(void)data;
	if (!strcmp(res->onoff, "on"))
		xbee_debug = 1;
	else
		xbee_debug = 0;
}

const char PROGMEM str_debug[] = "debug";
const char PROGMEM str_debug_onoff[] = "on#off";

const parse_token_string_t PROGMEM cmd_debug_debug =
	TOKEN_STRING_INITIALIZER(struct cmd_debug_result, debug,
				 str_debug);

const parse_token_string_t PROGMEM cmd_debug_onoff =
	TOKEN_STRING_INITIALIZER(struct cmd_debug_result, onoff,
				 str_debug_onoff);

const char PROGMEM help_debug[] = "enable/disable additionnal debug";

const parse_inst_t PROGMEM cmd_debug = {
	.f = cmd_debug_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_debug,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_debug_debug,
		(PGM_P)&cmd_debug_onoff,
		NULL,
	},
};

/**********************************************************/

/* this structure is filled when cmd_baudrate is parsed successfully */
struct cmd_baudrate_result {
	fixed_string_t arg0;
	uint32_t arg1;
};

/* function called when cmd_baudrate is parsed successfully */
static void cmd_baudrate_parsed(void * parsed_result, __attribute__((unused)) void *data)
{
	struct cmd_baudrate_result *res = parsed_result;
	struct uart_config c;

	uart_getconf(XBEE_UART, &c);
	c.baudrate = res->arg1;
	uart_setconf(XBEE_UART, &c);
}

const char PROGMEM str_baudrate_arg0[] = "baudrate";
const parse_token_string_t PROGMEM cmd_baudrate_arg0 =
	TOKEN_STRING_INITIALIZER(struct cmd_baudrate_result, arg0,
				 str_baudrate_arg0);
const parse_token_num_t PROGMEM cmd_baudrate_arg1 =
	TOKEN_NUM_INITIALIZER(struct cmd_baudrate_result, arg1,
			      UINT32);

const char PROGMEM help_baudrate[] = "Change xbee baudrate";
const parse_inst_t PROGMEM cmd_baudrate = {
	.f = cmd_baudrate_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_baudrate,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_baudrate_arg0,
		(PGM_P)&cmd_baudrate_arg1,
		NULL,
	},
};


/**********************************************************/

/* this structure is filled when cmd_beep is parsed successfully */
struct cmd_beep_result {
	fixed_string_t beep;
};

/* function called when cmd_beep is parsed successfully */
static void cmd_beep_parsed(void *parsed_result, void *data)
{
	(void)parsed_result;
	(void)data;

	beep(0, 3, 3);
	beep(1, 3, 3);
	beep(2, 3, 3);
	beep(0, 1, 1);
	beep(1, 1, 1);
	beep(2, 1, 1);
}

const char PROGMEM str_beep[] = "beep";
const parse_token_string_t PROGMEM cmd_beep_beep =
	TOKEN_STRING_INITIALIZER(struct cmd_beep_result, beep,
				 str_beep);

const char PROGMEM help_beep[] = "Send a beep";

const parse_inst_t PROGMEM cmd_beep = {
	.f = cmd_beep_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_beep,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_beep_beep,
		NULL,
	},
};

/**********************************************************/

/* this structure is filled when cmd_servo is parsed successfully */
struct cmd_servo_result {
	fixed_string_t arg0;
	fixed_string_t arg1;
	uint16_t num;
	uint16_t val;
};

/* function called when cmd_servo is parsed successfully */
static void cmd_servo_parsed(void * parsed_result, void *data)
{
	struct cmd_servo_result *res = parsed_result;

	(void)data;

	if (!strcmp_P(res->arg1, PSTR("set"))) {
		if (res->num >= N_SERVO) {
			printf_P(PSTR("bad servo num\n"));
			return;
		}
		if (res->val >= 1024) {
			printf_P(PSTR("bad servo val\n"));
			return;
		}
		spi_servo_set(res->num, res->val);
	}
	else if (!strcmp_P(res->arg1, PSTR("bypass"))) {
		spi_servo_set_bypass(!!res->val);
	}
	else if (!strcmp_P(res->arg1, PSTR("ppm"))) {
		spi_servo_set_ppm(!!res->val);
	}
	else if (!strcmp_P(res->arg1, PSTR("show"))) {
		spi_servo_dump();
	}
}

const char PROGMEM str_servo_arg0[] = "servo";
const parse_token_string_t PROGMEM cmd_servo_arg0 =
	TOKEN_STRING_INITIALIZER(struct cmd_servo_result, arg0,
				 str_servo_arg0);
const char PROGMEM str_servo_arg1_set[] = "set";
const parse_token_string_t PROGMEM cmd_servo_arg1_set =
	TOKEN_STRING_INITIALIZER(struct cmd_servo_result, arg1,
				 str_servo_arg1_set);
const parse_token_num_t PROGMEM cmd_servo_num =
	TOKEN_NUM_INITIALIZER(struct cmd_servo_result, num,
			      UINT16);
const parse_token_num_t PROGMEM cmd_servo_val =
	TOKEN_NUM_INITIALIZER(struct cmd_servo_result, val,
			      UINT16);

const char PROGMEM help_servo_set[] = "set servo value";
const parse_inst_t PROGMEM cmd_servo_set = {
	.f = cmd_servo_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_servo_set,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_servo_arg0,
		(PGM_P)&cmd_servo_arg1_set,
		(PGM_P)&cmd_servo_num,
		(PGM_P)&cmd_servo_val,
		NULL,
	},
};

const char PROGMEM str_servo_arg1_show[] = "show";
const parse_token_string_t PROGMEM cmd_servo_arg1_show =
	TOKEN_STRING_INITIALIZER(struct cmd_servo_result, arg1,
				 str_servo_arg1_show);

const char PROGMEM help_servo_show[] = "read servo and config";
const parse_inst_t PROGMEM cmd_servo_show = {
	.f = cmd_servo_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_servo_show,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_servo_arg0,
		(PGM_P)&cmd_servo_arg1_show,
		NULL,
	},
};

const char PROGMEM str_servo_arg1_bypassppm[] = "bypass#ppm";
const parse_token_string_t PROGMEM cmd_servo_arg1_bypassppm =
	TOKEN_STRING_INITIALIZER(struct cmd_servo_result, arg1,
				 str_servo_arg1_bypassppm);

const char PROGMEM help_servo_bypassppm[] = "change bypass/ppm";
const parse_inst_t PROGMEM cmd_servo_bypassppm = {
	.f = cmd_servo_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_servo_bypassppm,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_servo_arg0,
		(PGM_P)&cmd_servo_arg1_bypassppm,
		(PGM_P)&cmd_servo_val,
		NULL,
	},
};

/**********************************************************/

/* this structure is filled when cmd_test_spi is parsed successfully */
struct cmd_test_spi_result {
	fixed_string_t arg0;
};

static void cmd_test_spi_parsed(void * parsed_result, void *data)
{
	uint8_t i, flags, wait_time = 0;
	uint16_t val = 0;

	(void)parsed_result;
	(void)data;

	spi_servo_set_bypass(0);
	spi_servo_set_ppm(0);

	/* stress test: send many commands, no wait between each servo
	 * of a series, and a variable delay between series */
	printf_P(PSTR("stress test\r\n"));
	while (!cmdline_keypressed()) {

		wait_time++;
		if (wait_time > 20)
			wait_time = 0;

		IRQ_LOCK(flags);
		val = global_ms;
		IRQ_UNLOCK(flags);
		val >>= 3;
		val &= 1023;

		for (i = 0; i < 6; i++)
			spi_servo_set(i, val);

		wait_ms(wait_time);
		printf_P(PSTR("%4.4d %4.4d %4.4d %4.4d %4.4d %4.4d\r\n"),
			 spi_servo_get(0), spi_servo_get(1), spi_servo_get(2),
			 spi_servo_get(3), spi_servo_get(4), spi_servo_get(5));
	}

	printf_P(PSTR("bypass mode, with spi commands in background\r\n"));
	spi_servo_set_bypass(1);

	/* test bypass mode */
	while (!cmdline_keypressed()) {

		wait_time++;
		if (wait_time > 20)
			wait_time = 0;

		IRQ_LOCK(flags);
		val = global_ms;
		IRQ_UNLOCK(flags);
		val >>= 3;
		val &= 1023;

		for (i = 0; i < 6; i++)
			spi_servo_set(i, val);

		wait_ms(wait_time);
		printf_P(PSTR("%4.4d %4.4d %4.4d %4.4d %4.4d %4.4d\r\n"),
			 spi_servo_get(0), spi_servo_get(1), spi_servo_get(2),
			 spi_servo_get(3), spi_servo_get(4), spi_servo_get(5));
	}

	printf_P(PSTR("PPM to servo\r\n"));
	spi_servo_set_bypass(0);
	spi_servo_set_ppm(0);

	/* test PPM to servo (bypass) mode */
	while (!cmdline_keypressed()) {
		for (i = 0; i < 6; i++) {
			val = spi_servo_get(i);
			spi_servo_set(i, val);
		}
	}

	printf_P(PSTR("PPM to (servo + PPM)\r\n"));
	spi_servo_set_bypass(0);
	spi_servo_set_ppm(1);

	/* test PPM to servo (bypass) mode */
	while (!cmdline_keypressed()) {
		for (i = 0; i < 6; i++) {
			val = spi_servo_get(i);
			spi_servo_set(i, val);
		}
	}
}

const char PROGMEM str_test_spi_arg0[] = "test_spi";
const parse_token_string_t PROGMEM cmd_test_spi_arg0 =
	TOKEN_STRING_INITIALIZER(struct cmd_test_spi_result, arg0,
				 str_test_spi_arg0);

const char PROGMEM help_test_spi[] = "Test the spi";
const parse_inst_t PROGMEM cmd_test_spi = {
	.f = cmd_test_spi_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_test_spi,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_test_spi_arg0,
		NULL,
	},
};

/**********************************************************/

/* this structure is filled when cmd_dump_xbee_stats is parsed successfully */
struct cmd_dump_xbee_stats_result {
	fixed_string_t arg0;
};

static void cmd_dump_xbee_stats_parsed(void *parsed_result, void *data)
{
	(void)parsed_result;
	(void)data;

	xbee_dump_stats(xbee_dev);
}

const char PROGMEM str_dump_xbee_stats_arg0[] = "dump_xbee_stats";
const parse_token_string_t PROGMEM cmd_dump_xbee_stats_arg0 =
	TOKEN_STRING_INITIALIZER(struct cmd_dump_xbee_stats_result, arg0,
				 str_dump_xbee_stats_arg0);

const char PROGMEM help_dump_xbee_stats[] = "Test the spi";
const parse_inst_t PROGMEM cmd_dump_xbee_stats = {
	.f = cmd_dump_xbee_stats_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_dump_xbee_stats,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_dump_xbee_stats_arg0,
		NULL,
	},
};

/**********************************************************/

/* this structure is filled when cmd_test_eeprom_config is parsed successfully */
struct cmd_test_eeprom_config_result {
	fixed_string_t arg0;
};

static void cmd_test_eeprom_config_parsed(void *parsed_result, void *data)
{
	(void)parsed_result;
	(void)data;

	eeprom_dump_cmds();
	eeprom_append_cmd("salut1\n");
	eeprom_dump_cmds();
	eeprom_append_cmd("salut2\n");
	eeprom_append_cmd("salut3\n");
	eeprom_append_cmd("salut4\n");
	eeprom_dump_cmds();
	eeprom_insert_cmd_before("coin\n", 0);
	eeprom_insert_cmd_before("coin2\n", 2);
	eeprom_dump_cmds();
	eeprom_delete_cmd(2);
	eeprom_delete_cmd(0);
	eeprom_dump_cmds();
}

const char PROGMEM str_test_eeprom_config_arg0[] = "test_eeprom_config";
const parse_token_string_t PROGMEM cmd_test_eeprom_config_arg0 =
	TOKEN_STRING_INITIALIZER(struct cmd_test_eeprom_config_result, arg0,
				 str_test_eeprom_config_arg0);

const char PROGMEM help_test_eeprom_config[] = "Test the eeprom configuration";
const parse_inst_t PROGMEM cmd_test_eeprom_config = {
	.f = cmd_test_eeprom_config_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_test_eeprom_config,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_test_eeprom_config_arg0,
		NULL,
	},
};

/* ************* */

struct cmd_eeprom_del_result {
	fixed_string_t cmd;
	fixed_string_t action;
	uint8_t n;
};

static void cmd_eeprom_del_parsed(void *parsed_result,
				void *data)
{
	struct cmd_eeprom_del_result *res = parsed_result;

	(void)data;
	if (eeprom_delete_cmd(res->n) < 0)
		printf_P(PSTR("cannot delete command\n"));
	eeprom_dump_cmds();
}

const char PROGMEM str_eeprom_del_eeprom[] = "eeprom";
const parse_token_string_t PROGMEM cmd_eeprom_del_cmd =
	TOKEN_STRING_INITIALIZER(struct cmd_eeprom_del_result, cmd,
				 str_eeprom_del_eeprom);
const char PROGMEM str_eeprom_del_del[] = "del";
const parse_token_string_t PROGMEM cmd_eeprom_del_action =
	TOKEN_STRING_INITIALIZER(struct cmd_eeprom_del_result, action,
				 str_eeprom_del_del);
const parse_token_num_t PROGMEM cmd_eeprom_del_num =
	TOKEN_NUM_INITIALIZER(struct cmd_eeprom_del_result, n,
			      UINT8);

const char PROGMEM help_eeprom_del[] = "delete an eeprom init command";
const parse_inst_t PROGMEM cmd_eeprom_del = {
	.f = cmd_eeprom_del_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_eeprom_del,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_eeprom_del_cmd,
		(PGM_P)&cmd_eeprom_del_action,
		(PGM_P)&cmd_eeprom_del_num,
		NULL,
	},
};

/* ************* */

struct cmd_eeprom_add_result {
	fixed_string_t cmd;
	fixed_string_t action;
	uint8_t n;
};

static void cmd_eeprom_add_parsed(void *parsed_result,
				 void *data)
{
	struct cmd_eeprom_add_result *res = parsed_result;
	struct rdline rdl;
	const char *buffer;
	int8_t ret;
	int16_t c;

	rdline_init(&rdl, cmdline_write_char, NULL, NULL);
	rdline_newline(&rdl, "> ");

	/* XXX bad: we should not block as we do not serve callout */
	while (1) {
		c = cmdline_dev_recv(NULL);
		if (c < 0)
			continue;

		ret = rdline_char_in(&rdl, c);
		if (ret == -2) {
			printf_P(PSTR("abort\n"));
			return;
		}
		if (ret == 1)
			break;
	}

	buffer = rdline_get_buffer(&rdl);
	if (data == NULL)
		eeprom_insert_cmd_before(buffer, res->n);
	else
		eeprom_append_cmd(buffer);
	eeprom_dump_cmds();
}

const char PROGMEM str_eeprom_add_eeprom[] = "eeprom";
const parse_token_string_t PROGMEM cmd_eeprom_add_cmd =
	TOKEN_STRING_INITIALIZER(struct cmd_eeprom_add_result, cmd,
				 str_eeprom_add_eeprom);
const char PROGMEM str_eeprom_add_add[] = "add";
const parse_token_string_t PROGMEM cmd_eeprom_add_action =
	TOKEN_STRING_INITIALIZER(struct cmd_eeprom_add_result, action,
				 str_eeprom_add_add);
const parse_token_num_t PROGMEM cmd_eeprom_add_num =
	TOKEN_NUM_INITIALIZER(struct cmd_eeprom_add_result, n,
			      UINT8);

const char PROGMEM help_eeprom_add[] = "insert an eeprom init command";
const parse_inst_t PROGMEM cmd_eeprom_add = {
	.f = cmd_eeprom_add_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_eeprom_add,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_eeprom_add_cmd,
		(PGM_P)&cmd_eeprom_add_action,
		(PGM_P)&cmd_eeprom_add_num,
		NULL,
	},
};

const char PROGMEM help_eeprom_add2[] = "append an eeprom init command";
const parse_inst_t PROGMEM cmd_eeprom_add2 = {
	.f = cmd_eeprom_add_parsed,  /* function to call */
	.data = (void *)1,      /* 2nd arg of func */
	.help_str = help_eeprom_add2,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_eeprom_add_cmd,
		(PGM_P)&cmd_eeprom_add_action,
		NULL,
	},
};

/* ************* */

struct cmd_eeprom_list_result {
	fixed_string_t cmd;
	fixed_string_t action;
};

static void cmd_eeprom_list_parsed(void *parsed_result,
				void *data)
{
	(void)parsed_result;
	(void)data;
	eeprom_dump_cmds();
}

const char PROGMEM str_eeprom_list_eeprom[] = "eeprom";
const parse_token_string_t PROGMEM cmd_eeprom_list_cmd =
	TOKEN_STRING_INITIALIZER(struct cmd_eeprom_list_result, cmd,
				 str_eeprom_list_eeprom);
const char PROGMEM str_eeprom_list_list[] = "list";
const parse_token_string_t PROGMEM cmd_eeprom_list_action =
	TOKEN_STRING_INITIALIZER(struct cmd_eeprom_list_result, action,
				 str_eeprom_list_list);

const char PROGMEM help_eeprom_list[] = "list all eeprom init commands";
const parse_inst_t PROGMEM cmd_eeprom_list = {
	.f = cmd_eeprom_list_parsed,  /* function to call */
	.data = NULL,      /* 2nd arg of func */
	.help_str = help_eeprom_list,
	.tokens = {        /* token list, NULL terminated */
		(PGM_P)&cmd_eeprom_list_cmd,
		(PGM_P)&cmd_eeprom_list_action,
		NULL,
	},
};


/* ************* */

/* in progmem */
const parse_ctx_t PROGMEM main_ctx[] = {

	/* commands_gen.c */
	&cmd_reset,
	&cmd_bootloader,
	&cmd_log,
	&cmd_log_show,
	&cmd_log_type,
	&cmd_stack_space,
	&cmd_scheduler,
	&cmd_help,
	&cmd_neigh_del,
	&cmd_neigh_add,
	&cmd_neigh_list,
	&cmd_read,
	&cmd_write_none,
	&cmd_write_u8,
	&cmd_write_u16,
	&cmd_write_u32,
	&cmd_sendmsg,
	&cmd_send_hello,
	&cmd_send_hello_name,
	&cmd_sendmsg_name,
	&cmd_range,
	&cmd_range_period,
	&cmd_range_count,
	&cmd_range_powermask,
	&cmd_range_dstaddr,
	&cmd_monitor,
	&cmd_monitor_period,
	&cmd_monitor_add,
	&cmd_monitor_del,
	&cmd_ping,
	&cmd_raw,
	&cmd_dump,
	&cmd_debug,
	&cmd_baudrate,
	&cmd_beep,
	&cmd_servo_set,
	&cmd_servo_bypassppm,
	&cmd_servo_show,
	&cmd_test_spi,
	&cmd_dump_xbee_stats,
	&cmd_test_eeprom_config,
	&cmd_eeprom_del,
	&cmd_eeprom_add,
	&cmd_eeprom_add2,
	&cmd_eeprom_list,
	NULL,
};
