/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>

#include <aversive.h>
#include <aversive/queue.h>

#include <stdint.h>

#include <uart.h>

#include <scheduler.h>
#include <clock_time.h>
#include <parse.h>
#include <rdline.h>
#include <timer.h>
#include <xbee.h>

#include "callout.h"
#include "rc_proto.h"
#include "main.h"

/* */
struct rc_proto_power_levels {
	uint8_t ttl;
	uint16_t power_db;
};
static struct rc_proto_power_levels power_levels[MAX_POWER_LEVEL];

/* update power level when we receive the answer from DB */
static int update_power_level(void *frame, unsigned len, void *arg)
{
	struct xbee_atresp_hdr *atresp = (struct xbee_atresp_hdr *)frame;
	int level = (intptr_t)arg;
	uint8_t db;

	/* XXX check if this test is correct */
	if (len < sizeof(struct xbee_atresp_hdr) + sizeof(uint8_t)) {
		/* XXX stats */
		return -1;
	}

	db = atresp->data[0];
	power_levels[level].power_db = db;
	power_levels[level].ttl = 2;
	return 0;
}

/* when we receive a power probe, ask the DB value to the xbee */
void rc_proto_rx_power_probe(int power_level)
{
	(void)power_level;
	xbeeapp_send_atcmd("DB", NULL, 0, 0, update_power_level, NULL);
}

/* send a hello message */
// XXX iovec for xbee ?
int8_t rc_proto_send_hello(uint64_t addr, void *data, uint8_t data_len)
{
	struct {
		struct rc_proto_echo_req hdr;
		char buf[XBEE_MAX_FRAME_LEN - sizeof(struct rc_proto_echo_req)];
	} frame;

	if (data_len > sizeof(frame.buf))
		return -1;

	frame.hdr.type = RC_PROTO_HELLO;
	frame.hdr.datalen = data_len;
	memcpy(frame.buf, data, data_len);
	return xbeeapp_send_msg(addr, &frame,
				data_len + sizeof(struct rc_proto_echo_req), 1);
}


#if 0
#define N_SERVO 6
#define SERVO_NBITS 10
uint16_t servos[N_SERVO] = { 0x123, 0x234, 0x321, 0x123, 0x234, 0x321 };

int8_t servo2buf(uint8_t *buf, uint8_t len, uint8_t servo_mask,
		  uint8_t pow, uint8_t seq)
{
	uint8_t i = 0, num;
	uint8_t remain_bits;
	uint8_t servo_bits = 0;
	uint16_t servo_val;
	uint8_t tmp;

	if (len < 2)
		return -1;

	memset(buf, 0, len);
	buf[i++] = servo_mask;
	buf[i++] = ((seq & 0x1f) << 3) | (pow & 0x7);
	remain_bits = 8;

	for (num = 0; num < N_SERVO; num++) {
		if (!(servo_mask & (1 << num)))
			continue;
		servo_val = servos[num];
		servo_bits = SERVO_NBITS;

		tmp = (servo_val >> (servo_bits - remain_bits));
		tmp &= ((1 << remain_bits) - 1);
		if (i >= len)
			return -1;
		buf[i++] |= tmp;

		servo_bits = 10 - remain_bits;
		tmp = servo_val & ((1 << servo_bits) - 1);
		tmp <<= (8 - servo_bits);
		if (i >= len)
			return -1;
		buf[i] = tmp;

		if (servo_bits == 8) {
			i++;
			remain_bits = 8;
		}
		else
			remain_bits = 8 - servo_bits;
	}

	if (remain_bits != 8)
		i++;

	return i;
}

int8_t buf2servo(uint8_t *buf, uint8_t len)
{
	uint8_t mask, count = 0;
	uint8_t num = 0;
	uint8_t pow, seq;
	uint16_t val;

	if (len < 2)
		return -1;

	mask = buf[0];
	if (mask > 0x3f)
		return -1;
	pow = buf[1] & 0x07;
	seq = buf[1] >> 5;

	for (num = 0; num < N_SERVO; num++) {
		if ((1<<num) & mask)
			count++;
	}
	switch (count) {
		case 1: if (len != 4) return -1; break;
		case 2: if (len != 5) return -1; break;
		case 3: if (len != 6) return -1; break;
		case 4: if (len != 7) return -1; break;
		case 5: if (len != 9) return -1; break;
		case 6: if (len != 10) return -1; break;
		default: return -1;
	}

	for (num = 0; ((1<<num) & mask) == 0; num++) {
		if (num >= N_SERVO)
			return 0;
	}

	val = buf[2];
	val <<= 2;
	val |= (buf[3] >> 6);

	for (num++; ((1<<num) & mask) == 0; num++) {
		if (num >= N_SERVO)
			return 0;
	}

	val = buf[3] & 0x3f;
	val <<= 4;
	val |= (buf[4] >> 4);

	for (num++; ((1<<num) & mask) == 0; num++) {
		if (num >= N_SERVO)
			return 0;
	}

	val = buf[4] & 0xf;
	val <<= 6;
	val |= (buf[5] >> 2);

	for (num++; ((1<<num) & mask) == 0; num++) {
		if (num >= N_SERVO)
			return 0;
	}

	val = buf[5] & 0x3;
	val <<= 8;
	val |= (buf[6]);

	for (num++; ((1<<num) & mask) == 0; num++) {
		if (num >= N_SERVO)
			return 0;
	}

	val = buf[7];
	val <<= 2;
	val |= (buf[8] >> 6);

	for (num++; ((1<<num) & mask) == 0; num++) {
		if (num >= N_SERVO)
			return 0;
	}

	val = buf[8];
	val <<= 4;
	val |= (buf[9] >> 4);

	return 0;
}
#endif
